from json.decoder import JSONDecodeError
import pandas as pd
from .EuroLeagueData import EuroLeagueData
from .boxscore_data import BoxScoreData
from .utils import get_requests


class PlayByPlay(EuroLeagueData):
    """
    A class for getting the game play-by-play data.

    Args:
        competition (str, optional): The competition code, inherited from the
            `EuroLeagueData` class. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """

    def get_game_play_by_play_data(
        self,
        season: int,
        gamecode: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data of a particular game.

        Args:

            season (int): The start year of the season

            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.

        Returns:

            pd.DataFrame: A dataframe with the play-by-play data of the game.
        """
        url = "https://live.euroleague.net/api/PlaybyPlay"
        params = {
            "gamecode": gamecode,
            "seasoncode": f"{self.competition}{season}"
        }
        r = get_requests(url, params=params)

        try:
            data = r.json()
        except JSONDecodeError:
            raise ValueError(
                f"Game code, {gamecode}, did not return any data.")

        periods = [
            'FirstQuarter', 'SecondQuarter', 'ThirdQuarter', 'ForthQuarter',
            'ExtraTime'
        ]
        all_data = []
        for p, period in enumerate(periods):
            if data[period]:
                df = pd.json_normalize(data[period])
                df["PERIOD"] = p + 1
                all_data.append(df)

        play_by_play_df = pd.concat(all_data).reset_index(drop=True)
        play_by_play_df['CODETEAM'] = play_by_play_df['CODETEAM'].str.strip()
        play_by_play_df['PLAYER_ID'] = play_by_play_df['PLAYER_ID'].str.strip()
        play_by_play_df.insert(0, 'Season', season)
        play_by_play_df.insert(1, 'Gamecode', gamecode)
        return play_by_play_df

    def get_game_play_by_play_data_single_season(
        self,
        season: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data of *all* games in a single
        season

        Args:

            season (int): The start year of the season

        Returns:

            pd.DataFrame: A dataframe with the play-by-play data of all games
                in a single season
        """
        data_df = self.get_season_data_from_game_data(
            season, self.get_game_play_by_play_data)
        return data_df

    def get_game_play_by_play_data_multiple_seasons(
        self, start_season: int, end_season: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data of *all* games in a range of
        seasons

        Args:

            start_season (int): The start year of the start season

            end_season (int): The start year of the end season

        Returns:

            pd.DataFrame: A dataframe with the play-by-play data of all games
                in range of seasons
        """
        df = self.get_range_seasons_data(
            start_season, end_season, self.get_game_play_by_play_data)
        return df

    def get_lineups_data(self, season, gamecode):
        """
        Get the teams' lineups and the minute of the game these lineups
        change.

        Args:

            season (int): The start year of the season

            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.

        Returns:

            pd.DataFrame: A dataframe with the lineups of the teams as they
                change during the game (minute of change is provided)
        """
        # Get the starting line-ups from boxscore data
        boxscoredata = BoxScoreData(competition=self.competition)
        game_bxscr_stats = boxscoredata.get_player_boxscore_stats_data(
            season=season, gamecode=gamecode)

        # find home and away teams
        hm_aw = game_bxscr_stats[["Home", "Team"]].drop_duplicates()
        home_team = hm_aw.loc[hm_aw["Home"] == 1, "Team"].values[0]
        away_team = hm_aw.loc[hm_aw["Home"] == 0, "Team"].values[0]

        starting_five = game_bxscr_stats.loc[
            game_bxscr_stats["IsStarter"] == 1, ["Team", "Player"]
        ]
        starting_five['ID'] = starting_five.groupby('Team').cumcount()

        # Pivot the DataFrame
        starting_five = starting_five.pivot(
            index='ID', columns='Team', values='Player')

        # Reset index if needed
        starting_five.reset_index(drop=True, inplace=True)
        starting_five = starting_five[[home_team, away_team]]
        starting_five_dict = starting_five.to_dict(orient='list')

        # Build the dataframe, whose first row is the starting lineups
        df = pd.DataFrame(
            {
                "Season": [season],
                "Gamecode": [gamecode],
                "PERIOD": [1],
                "MARKERTIME": ["10:00"],
                "MINUTE": [1],
                f"Lineup_{home_team}": [starting_five_dict[home_team]],
                f"Lineup_{away_team}": [starting_five_dict[away_team]],
            }
        )

        # Fetch play-by-play data
        pbp_data = self.get_game_play_by_play_data(
            season=season, gamecode=gamecode)

        # focus on the subs entries
        subs_mask = pbp_data["PLAYTYPE"].isin(["IN", "OUT"])
        subs_data = pbp_data[subs_mask].reset_index(drop=True)

        # pivot to create a dataframe with in/out player by minute and team
        #  side by side, i.e. column-wise.
        subs_data_pivot = subs_data.pivot_table(
            index=["PERIOD", "MINUTE", "MARKERTIME", "CODETEAM"],
            columns="PLAYTYPE",
            values="PLAYER",
            aggfunc=list
        )

        # iterate this dataframe and add/remove players from the line-up list
        n = 0
        home_existing_lineup = starting_five_dict[home_team]
        away_existing_lineup = starting_five_dict[away_team]
        for r, row in subs_data_pivot.iterrows():
            period = r[0]
            minute = r[1]
            time = r[2]
            team = r[3]
            sub_d = dict(zip(row["OUT"], row["IN"]))

            n += 1
            if team == home_team:
                home_new_lineup = [
                    sub_d[u] if u in sub_d else u
                    for u in home_existing_lineup
                ]
                home_existing_lineup = home_new_lineup.copy()

                df.loc[n] = [season, gamecode, period, time,
                             minute, home_new_lineup, away_existing_lineup]
            else:
                away_new_lineup = [
                    sub_d[u] if u in sub_d else u
                    for u in away_existing_lineup
                ]
                away_existing_lineup = away_new_lineup.copy()

                df.loc[n] = [season, gamecode, period, time,
                             minute, home_existing_lineup, away_new_lineup]

        cols = ["Season", "Gamecode", "PERIOD", "MARKERTIME", "MINUTE"]
        # If the two teams made subs in the same minute, these would have
        # been recorded in two seperate lines in the for loop above.
        # Fix the unique minute and keep the latest values of the subs lists.
        line_ups_df = df.drop_duplicates(
            cols, keep="last").reset_index(drop=True)

        return line_ups_df

    def get_pbp_data_with_lineups(self, season, gamecode):
        """
        Get the play-by-play data enrighed with the teams lineups for
        every minute in the PBP data.

        Args:

            season (int): The start year of the season

            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.

        Returns:

            pd.DataFrame: A dataframe with the play-by-play enriched with
                teams' line ups
        """
        pbp_data = self.get_game_play_by_play_data(
            season=season, gamecode=gamecode)
        line_ups_df = self.get_lineups_data(
            season=season, gamecode=gamecode)
        # hack for consistency
        pbp_data.loc[pbp_data["PLAYTYPE"] == "BP", "MARKERTIME"] = "10:00"

        # cols = ["Season", "Gamecode", "PERIOD", "MARKERTIME", "MINUTE"]
        # lineup_cols = [f"Lineup_{home_team}", f"Lineup_{away_team}"]
        cols = list(line_ups_df.columns[:5])
        lineup_cols = list(line_ups_df.columns[-2:])
        # merge with PBP dataframe
        pbp_lu_data = pbp_data.merge(line_ups_df, on=cols, how="left")
        # update the starting lineups
        pbp_lu_data.loc[0, lineup_cols] = line_ups_df.loc[0, lineup_cols]
        # forward fill the lineups until a lineup entry changes
        pbp_lu_data[lineup_cols] = pbp_lu_data[lineup_cols].ffill()
        # fix hack
        pbp_lu_data.loc[pbp_lu_data["PLAYTYPE"] == "BP", "MARKERTIME"] = ""

        return pbp_lu_data



from typing import Optional, List
import requests


def get_requests(
    url: str,
    params: dict = {},
    headers: dict = {"Accept": "application/json"}
) -> requests.models.Response:
    """
    A wrapper to `requests.get()` which handles unsuccesful requests too.

    Args:

        url (str): _description_

        params (dict, optional): The `params` variables in get requests.
            Defaults to {}.

        headers (dict, optional): the `header` variable in get requests.
            Defaults to {"Accept": "application/json"}.

    Raises:

        Requests Error: If get request was not succesful

    Returns:

        requests.models.Response: The response object.
    """
    r = requests.get(url, params=params, headers=headers, timeout=60)

    if r.status_code != 200:
        r.raise_for_status()

    return r


def raise_error(
    var: Optional[str],
    descripitve_var: str,
    available_vals: List,
    allow_none: bool = False,
) -> None:
    """
    A function that raises a ValueError with specific message.

    Args:

        var (str): The input variable by the user

        descripitve_var (str): A description in plain English of this variable

        available_vals (List): The available variables

        allow_none (bool, optional): If `var` can take None value.
            Defaults to False.

    Raises:

        ValueError: if `var` not applicable
    """

    if allow_none:
        available_vals.append(None)

    if var not in available_vals:
        raise ValueError(
            f"{descripitve_var}, {var}, is not applicable. "
            f"Available values: {available_vals}"
        )
    return



import pandas as pd
from .EuroLeagueData import EuroLeagueData
from .utils import (
    raise_error,
    get_requests
)


class GameStats(EuroLeagueData):
    """
    A class for getting the game related stats and data.

    Args:
        competition (str, optional): The competition code, inherited from the
            `EuroLeagueData` class. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """

    def get_game_data(
        self,
        season: int,
        game_code: int,
        endpoint: str
    ) -> pd.DataFrame:
        """
        A wrapper function for getting game-level data.

        Args:

            season (int): The start year of the season

            game_code (int): The game code of the game of interest.
                Find the game code from Euroleague's website

            endpoint (str): The type of game data, available variables:
                - report
                - stats
                - teamsComparison

        Raises:

            ValueError: If input endpoint is not applicable.

        Returns:

            pd.DataFrame: A dataframe with the game data.
        """
        game_endpoints = ["report", "stats", "teamsComparison"]

        raise_error(endpoint, "Statistic type", game_endpoints, False)

        url_ = self.make_season_game_url(season, game_code, endpoint)
        r = get_requests(url_)

        data = r.json()
        df = pd.json_normalize(data)
        df.insert(0, "Season", season)
        if "gameCode" in df.columns:
            df.rename(columns={"gameCode": "Gamecode"}, inplace=True)
        else:
            df.insert(1, "Gamecode", game_code)
        if "round" in df.columns:
            df.rename(columns={"round": "Round"}, inplace=True)

        return df

    def get_game_report(self, season: int, game_code: int) -> pd.DataFrame:
        """
        Get game report data for a game

        Args:
            season (int): The start year of the season
            game_code (int): The game code of the game of interest. It can be
                found on Euroleague's website.

        Returns:
            pd.DataFrame: A dataframw with the game report data
        """
        df = self.get_game_data(season, game_code, "report")
        return df

    def get_game_reports_single_season(self, season: int) -> pd.DataFrame:
        """
        Get game report data for *all* games in a single season

        Args:
            season (int): The start year of the season

        Returns:
            pd.DataFrame: A dataframe with game report data
        """
        data_df = self.get_season_data_from_game_data(
            season, self.get_game_report)
        return data_df

    def get_game_reports_range_seasons(
        self,
        start_season: int,
        end_season: int
    ) -> pd.DataFrame:
        """
        Get game report data for *all* games in a range of seasons

        Args:

            start_season (int): The start year of the start season

            end_season (int): The start year of the end season

        Returns:

            pd.DataFrame: A dataframe with game report data
        """
        df = self.get_range_seasons_data(
            start_season, end_season, self.get_game_report)
        return df

    def get_game_stats(self, season: int, game_code: int) -> pd.DataFrame:
        """
        Get game stats data for single game

        Args:

            season (int): The start year of the season

            game_code (int): The game code of the game of interest. It can be
                found on Euroleague's website.

        Returns:

            pd.DataFrame: A dataframe with the games' stats data
        """
        df = self.get_game_data(season, game_code, "stats")
        return df

    def get_game_stats_single_season(self, season: int) -> pd.DataFrame:
        """
        Get game stats data for *all* games in a single season

        Args:

            season (int): The start year of the season

        Returns:

            pd.DataFrame: A dataframe with the games' stats data
        """
        data_df = self.get_season_data_from_game_data(
            season, self.get_game_stats)
        return data_df

    def get_game_stats_range_seasons(
        self,
        start_season: int,
        end_season: int
    ) -> pd.DataFrame:
        """
        Get game stats data for *all* games in a range of seasons

        Args:

            start_season (int): The start year of the start season

            end_season (int): The start year of the end season

        Returns:

            pd.DataFrame: A dataframe with the games' stats data
        """
        df = self.get_range_seasons_data(
            start_season, end_season, self.get_game_stats)
        return df

    def get_game_teams_comparison(
        self,
        season: int,
        game_code: int
    ) -> pd.DataFrame:
        """
        A function that gets the "teams comparison" game stats for a single
        game. This is the *pre-game* stats. Hence gamecodes of round 1 of each
        season are not available.

        Args:

            season (int): The start year of the season

            game_code (int): The game code of the game of interest. It can be
                found on Euroleague's website.

        Returns:

            pd.DataFrame: A dataframe with games teams comparison stats
        """
        df = self.get_game_data(season, game_code, "teamsComparison")
        return df

    def get_game_teams_comparison_single_season(
            self, season: int) -> pd.DataFrame:
        """
        A function that gets the pre-grame "teams comparison" game stats for
        *all* games in a single season.

        Args:

            season (int): The start year of the season

        Returns:

            pd.DataFrame: A dataframe with games teams comparison stats
        """
        data_df = self.get_season_data_from_game_data(
            season, self.get_game_teams_comparison)
        return data_df

    def get_game_teams_comparison_range_seasons(
        self,
        start_season: int,
        end_season: int
    ) -> pd.DataFrame:
        """
        A function that gets the pre-game "teams comparison" game stats for
        *all* in a range seasons

        Args:

            start_season (int): The start year of the star season

            end_season (int): The start year of the end season

        Returns:

            pd.DataFrame: A dataframe with games teams comparison stats
        """
        df = self.get_range_seasons_data(
            start_season,
            end_season,
            self.get_game_teams_comparison
        )
        return df



from typing import Optional
import pandas as pd
import numpy as np
from .EuroLeagueData import EuroLeagueData
from .boxscore_data import BoxScoreData
from .utils import (
    raise_error,
    get_requests
)


class TeamStats(EuroLeagueData):
    """
    A class for getting team-level stats and data.

    Args:
        competition (str, optional): The competition code, inherited from the
            `EuroLeagueData` class. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """

    def get_team_stats(
        self,
        endpoint: str,
        params: dict = {},
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame"
    ) -> pd.DataFrame:
        """
        A wrapper functions for collecting teams' stats.
        Allows for three types of data:
        - all seasons
        - single season
        - range of seasons

        Args:

            endpoint (str): The type of stats to fetch. Available values:
                - traditional
                - advanced
                - opponentsTraditional
                - opponentsAdvanced

            params (Dict[str, Union[str, int]]): A dictionary of the parmaters
                for the get request.

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                Defaults to "PerGame".

        Raises:

            ValueError: If the endpoint is not applicable

            ValueError: If the phase_type_code is not applicable

            ValueError: If the statistic_mode is not applicable

        Returns:

            pd.DataFrame: A dataframe with the teams' stats.
        """

        available_endpoints = [
            "traditional", "advanced",
            "opponentsTraditional", "opponentsAdvanced"
        ]
        available_phase_type_code = ["RS", "PO", "FF"]
        available_stat_mode = ["PerGame", "Accumulated"]

        raise_error(
            endpoint, "Statistic type", available_endpoints, False)
        raise_error(
            statistic_mode, "Statistic Aggregation", available_stat_mode,
            False)
        raise_error(
            phase_type_code, "Phase type code", available_phase_type_code,
            True)

        params["statisticMode"] = statistic_mode
        params["phaseTypeCode"] = phase_type_code
        params["limit"] = 400

        url_ = f"{self.url}/statistics/teams/{endpoint}"

        r = get_requests(url_, params=params)
        data = r.json()
        if data["total"] < len(data["teams"]):
            params["limit"] = len(data["teams"]) + 1
            r = get_requests(url_, params=params)
            data = r.json()
        df = pd.json_normalize(data["teams"])
        return df

    def get_team_stats_all_seasons(
        self,
        endpoint: str,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame"
    ) -> pd.DataFrame:
        """
        A function that gets the team stats for all seasons

        Args:

            endpoint (str): The type of stats to fetch. Available values:
                - traditional
                - advanced
                - opponentsTraditional
                - opponentsAdvanced

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                Defaults to "PerGame".

        Returns:

            pd.DataFrame: A dataframe with the teams' stats
        """
        params = {"SeasonMode": "All"}
        df = self.get_team_stats(
            endpoint, params, phase_type_code, statistic_mode)
        return df

    def get_team_stats_single_season(
        self,
        endpoint: str,
        season: int,
        phase_type_code: str,
        statistic_mode: str
    ) -> pd.DataFrame:
        """
        A function that returns the teams' stats in a single season

        Args:

            endpoint (str): The type of stats to fetch. Available values:
                - traditional
                - advanced
                - opponentsTraditional
                - opponentsAdvanced

            season (int): The start year of the season

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                Defaults to "PerGame".

        Returns:

            pd.DataFrame: A dataframe with the teams' stats
        """
        params = {
            "SeasonMode": "Single",
            "SeasonCode": f"{self.competition}{season}",
        }
        df = self.get_team_stats(
            endpoint, params, phase_type_code, statistic_mode)
        return df

    def get_team_stats_range_seasons(
        self,
        endpoint: str,
        start_season: int,
        end_season: int,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame"
    ) -> pd.DataFrame:
        """
        A function that returns the teams' stats in a range of seasons

        Args:

            endpoint (str): The type of stats to fetch. Available values:
                - traditional
                - advanced
                - opponentsTraditional
                - opponentsAdvanced

            start_season (int): The start year of the start season

            end_season (int): The end year of the end season

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                Defaults to "PerGame".

        Returns:

            pd.DataFrame: A dataframe with the teams' stats
        """
        params = {
            "SeasonMode": "Range",
            "FromSeasonCode": f"{self.competition}{start_season}",
            "ToSeasonCode": f"{self.competition}{end_season}",
        }
        df = self.get_team_stats(
            endpoint, params, phase_type_code, statistic_mode)
        return df

    def get_team_advanced_stats_single_game(self, season: int, gamecode: int):
        """
        In this function we derive team advanced stats from a single game
        that are not provided by the API but can be easily estimated from
        stats that are given from the API, i.e.
            - Number of Possessions
            - Pace
        The formulas and definitions of these stats can be found in
            - [Basketball-reference.com](https://www.basketball-reference.com/about/glossary.html)  # noqa
            - [kenpom](https://kenpom.com/blog/the-possession/)
            - [hackastat](https://hackastat.eu/en/learn-a-stat-possessions-and-pace/)  # noqa

        Args:
            season (int): The start year of the season
            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.
        """
        boxscoredata = BoxScoreData(competition=self.competition)
        game_bxscr_stats = boxscoredata.get_player_boxscore_stats_data(
            season=season, gamecode=gamecode)
        totals_df = (
            game_bxscr_stats[game_bxscr_stats["Player_ID"] == "Total"]
            .set_index("Team").T
        )
        fga = (totals_df.loc["FieldGoalsAttempted2"] +
               totals_df.loc["FieldGoalsAttempted3"])

        fgm = (totals_df.loc["FieldGoalsMade2"] +
               totals_df.loc["FieldGoalsMade3"])
        possessions_simple = (
            fga
            + 0.44 * (totals_df.loc["FreeThrowsAttempted"])
            - totals_df.loc["OffensiveRebounds"] + totals_df.loc["Turnovers"]
        )

        possessions = (
            fga
            + 0.44 * (totals_df.loc["FreeThrowsAttempted"])
            - 1.07 * (
                totals_df.loc["OffensiveRebounds"] /
                totals_df.loc["DefensiveRebounds", ::-1].values
            ) * (fga - fgm)
            + totals_df.loc["Turnovers"]
        )

        df = pd.DataFrame(
            {
                "Possessions (simple)": possessions_simple,
                "Possessions": possessions,
            }
        )
        df.loc["Game"] = [possessions_simple.mean(), possessions.mean()]

        min_played_ls = totals_df.loc["Minutes"][0].split(":")
        min_played = int(min_played_ls[0]) + float(min_played_ls[1]) / 60
        min_factor = (min_played / 5)

        df["Pace (simple)"] = [
            np.nan, np.nan, 40 * possessions_simple.mean() / min_factor
        ]
        df["Pace"] = [np.nan, np.nan, 40 * possessions.mean() / min_factor]
        df["Season"] = season
        df["Gamecode"] = gamecode
        df["Home"] = [1, 0, np.nan]

        df = df[
            [
                "Season", "Gamecode", "Home",
                "Possessions (simple)", "Possessions",
                "Pace (simple)", "Pace",
            ]
        ]
        return df



from typing import Callable
import logging
import pandas as pd
from tqdm.auto import tqdm, trange
from requests.exceptions import HTTPError
import xmltodict
from .utils import get_requests

logging.basicConfig(encoding='utf-8', level=logging.INFO)
logger = logging.getLogger(__name__)


class EuroLeagueData:
    """
    Base class for collecting Euroleague and Eurocup competition's data.

    Args:
        competition (str, optional): The competition code. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """
    BASE_URL = "https://api-live.euroleague.net"
    VERSION = "v3"

    def __init__(self, competition="E"):
        """_summary_

        Args:
            competition (str, optional): The competition code. Choose one of:
                - 'E' for Euroleague
                - 'U' for Eurocup
                Defaults to "E".

        Raises:
            ValueError: When an invalid competition code is provided.
        """
        if competition not in ["E", "U"]:
            raise ValueError(
                "Invalid competition value, {competition}. "
                "Valid values 'E', 'U'"
            )
        self.competition = competition
        self.url = f"{self.BASE_URL}/{self.VERSION}/competitions/{competition}"

    def make_season_game_url(
        self,
        season: int,
        game_code: int,
        endpoint: str
    ) -> str:
        """
        Concatenates the base URL and makes the game url.

        Args:

            season (int): The start year of the season

            game_code (int): The code of the game. Find the code from
                Euroleague's website

            endpoint (str): The endpoint of the API

        Returns:

            str: the full URL
        """
        full_url = (
            f"{self.url}/seasons/{self.competition}{season}/"
            f"games/{game_code}/{endpoint}"
        )
        return full_url

    def get_game_metadata_season(self, season: int) -> pd.DataFrame:
        """
        A function that returns the game metadata, e.g. gamecodes of season

        Args:

            season (int): The start year of the season.

        Returns:

            pd.DataFrame: A dataframe with the season's game metadata, e.g.
                gamecode, score, home-away teams, date, round, etc.
        """
        url = (
            "https://api-live.euroleague.net/v1/results"
            f"?seasonCode={self.competition}{season}"
        )
        r = get_requests(url)
        data = xmltodict.parse(r.content)
        df = pd.DataFrame(data["results"]["game"])
        int_cols = ["gameday", "gamenumber", "homescore", "awayscore"]
        df[int_cols] = df[int_cols].astype(int)
        df["played"] = df["played"].replace({"true": True, "false": False})
        return df

    def get_season_data_from_game_data(
        self,
        season: int,
        fun: Callable[[int, int], pd.DataFrame]
    ) -> pd.DataFrame:
        """
        A wrapper function for getting game data for all games in a single
        season.

        Args:

            season (int, optional): The start year of the season.

            fun (Callable[[int, int], pd.DataFrame]): A callable function that
                determines that type of data to be collected. Available values:
                - get_game_report
                - get_game_stats
                - get_game_teams_comparison

        Returns:

            pd.DataFrame: A dataframe with the game data.
        """
        data_list = []

        game_metadata_df = self.get_game_metadata_season(season)
        game_metadata_df = game_metadata_df[game_metadata_df["played"]]
        game_codes_df = (
            game_metadata_df[["round", "gameday", "gamenumber"]]
            .drop_duplicates().sort_values(["gamenumber", "gameday"])
            .reset_index(drop=True)
        )
        for _, row in tqdm(game_codes_df.iterrows(),
                           total=game_codes_df.shape[0],
                           desc=f"Season {season}", leave=True):
            game_code = row["gamenumber"]
            try:
                df = fun(season, game_code)
                if df.empty:
                    logger.warning(f"Game {game_code} returned no data.")
                    continue
                if ("Phase" not in df.columns) and ("round" in row):
                    df.insert(1, "Phase", row["round"])
                if ("Round" not in df.columns) and ("gameday" in row):
                    df.insert(2, "Round", row["gameday"])
                data_list.append(df)
            except HTTPError as err:
                logger.warning(
                    f"HTTPError: Didn't find gamecode {game_code} for season "
                    f"{season}. Invalid {err}. Skip and continue."
                )
            except:  # noqa: E722
                logger.warning(
                    f"Something went wrong for game {game_code}. "
                    "Skip and continue"
                )

        if data_list:
            data_df = pd.concat(data_list, axis=0)
            data_df.reset_index(drop=True, inplace=True)
        else:
            data_df = pd.DataFrame([])
        return data_df

    def get_range_seasons_data(
        self,
        start_season: int,
        end_season: int,
        fun: Callable[[int, int], pd.DataFrame]
    ) -> pd.DataFrame:
        """
        A wrapper function with the all game data in a range of seasons

        Args:

            season (int, optional): The start year of the season.

            end_season (int): The start year of teh end season

            fun (Callable[[int, int], pd.DataFrame]): A callable function that
                determines that type of data to be collected. Available values:
                - get_game_report
                - get_game_stats
                - get_game_teams_comparison

        Returns:

            pd.DataFrame: A dataframe with the game data
        """
        data = []
        for season in trange(
                start_season, end_season + 1, desc="Season loop", leave=True):

            data_df = self.get_season_data_from_game_data(season, fun)
            data.append(data_df)
        df = pd.concat(data)
        df.reset_index(drop=True, inplace=True)
        return df



from . import game_stats
from . import player_stats
from . import shot_data
from . import standings
from . import team_stats
from . import play_by_play_data
from . import boxscore_data
from . import utils

__all__ = [
    "game_stats",
    "player_stats",
    "shot_data",
    "standings",
    "team_stats",
    "play_by_play_data",
    "boxscore_data",
    "utils"
]



from json.decoder import JSONDecodeError
import pandas as pd
from .EuroLeagueData import EuroLeagueData
from .utils import get_requests


class ShotData(EuroLeagueData):
    """
    A class for getting shot data.

    Args:
        competition (str, optional): The competition code, inherited from the
            `EuroLeagueData` class. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """
    # MADE_ACTIONS = ['2FGM', '3FGM', 'LAYUPMD', 'DUNK']
    # MISSES_ACTIONS = ['2FGA', '2FGAB', '3FGA', '3FGAB', 'LAYUPATT']

    def get_game_shot_data(self, season: int, gamecode: int) -> pd.DataFrame:
        """
        A function that gets the shot data of a particular game.

        Args:

            season (int): The start year of the season

            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.

        Returns:

            pd.DataFrame: A dataframe with the shot data of the game.
        """
        url = "https://live.euroleague.net/api/Points"
        params = {
            "gamecode": gamecode,
            "seasoncode": f"{self.competition}{season}"
        }
        r = get_requests(url, params=params)

        try:
            data = r.json()
        except JSONDecodeError:
            raise ValueError(
                f"Game code, {gamecode}, did not return any data.")

        shots_df = pd.DataFrame(data['Rows'])
        # team id, player id and action id contain trailing white space
        if not shots_df.empty:
            shots_df['TEAM'] = shots_df['TEAM'].str.strip()
            shots_df['ID_PLAYER'] = shots_df['ID_PLAYER'].str.strip()
            shots_df['ID_ACTION'] = shots_df['ID_ACTION'].str.strip()
            shots_df.insert(0, 'Season', season)
            shots_df.insert(1, 'Gamecode', gamecode)
        return shots_df

    def get_game_shot_data_single_season(self, season: int) -> pd.DataFrame:
        """
        A function that gets the shot data of *all* games in a single season

        Args:

            season (int): The start year of the season

        Returns:

            pd.DataFrame: A dataframe with the shot data of all games in a
                single season
        """
        data_df = self.get_season_data_from_game_data(
            season, self.get_game_shot_data)
        return data_df

    def get_game_shot_data_multiple_seasons(
        self, start_season: int, end_season: int
    ) -> pd.DataFrame:
        """
        A function that gets the shot data of *all* games in a range of seasons

        Args:

            start_season (int): The start year of the start season

            end_season (int): The start year of the end season

        Returns:

            pd.DataFrame: A dataframe with the shot data of all games in range
                of seasons
        """
        df = self.get_range_seasons_data(
            start_season, end_season, self.get_game_shot_data)
        return df



from typing import List
from json.decoder import JSONDecodeError
import pandas as pd
from .EuroLeagueData import EuroLeagueData
from .utils import (
    get_requests,
    raise_error
)


class BoxScoreData(EuroLeagueData):
    """
    A class for getting box-score data

    Args:
        competition (str, optional): The competition code, inherited from the
            `EuroLeagueData` class. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """

    def get_boxscore_data(
        self,
        season: int,
        gamecode: int,
        boxscore_type: str = "ByQuarter"
    ) -> List[dict]:
        """A helper function that gets the boxscore data of a particular data.

        Args:
            season (int): The start year of the season
            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.
            boxscore_type (str, optional): The type of quarter boxscore data.
                Available values:
                - Stats
                - ByQuarter
                - EndOfQuarter
                Defaults to "ByQuarter".

        Raises:
            ValueError: If boxscore_type value is not valid.

        Returns:
            List[dict]: A list of dictionaries with the data.
        """
        url = "https://live.euroleague.net/api/Boxscore"
        params = {
            "gamecode": gamecode,
            "seasoncode": f"{self.competition}{season}"
        }
        r = get_requests(url, params=params)

        try:
            data = r.json()
        except JSONDecodeError:
            raise ValueError(
                f"Game code, {gamecode}, did not return any data.")
        boxscore_types = ["Stats", "ByQuarter", "EndOfQuarter"]

        if boxscore_type not in boxscore_types:
            raise_error(boxscore_type, "Boxscore", boxscore_types, False)

        return data[boxscore_type]

    def get_game_boxscore_quarter_data(
        self,
        season: int,
        gamecode: int,
        boxscore_type: str = "ByQuarter"
    ) -> pd.DataFrame:
        """
        A function that gets the boxscore quarterly data of a particular game.

        Args:

            season (int): The start year of the season

            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.

            boxscore_type (str): The type of quarter boxscore data.
                Available values:
                - ByQuarter
                - EndOfQuarter
                Default: ByQuarter

        Raises:
            ValueError: If boxscore_type value is not valid.

        Returns:

            pd.DataFrame: A dataframe with the boxscore quarter data of the
                game.
        """
        valid_vals = ["ByQuarter", "EndOfQuarter"]
        if boxscore_type not in valid_vals:
            raise_error(boxscore_type, "Boxscore quarter type",
                        valid_vals, False)

        data = self.get_boxscore_data(season, gamecode, boxscore_type)
        df = pd.json_normalize(data)
        df.insert(0, 'Season', season)
        df.insert(1, 'Gamecode', gamecode)
        return df

    def get_player_boxscore_stats_data(
        self,
        season: int,
        gamecode: int
    ) -> pd.DataFrame:
        """
        The players' and team's total stats of a particular game.

        Args:
            season (int): The start year of the season
            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.

        Returns:
            pd.DataFrame: A dataframe with home and away team player stats
        """
        data = self.get_boxscore_data(season, gamecode, "Stats")
        home_df = pd.concat([
            pd.json_normalize(data[0]["PlayersStats"]),
            pd.json_normalize(data[0]["tmr"]),
            pd.json_normalize(data[0]["totr"])
        ])
        home_df.reset_index(drop=True, inplace=True)
        home_df.iloc[-2:, 0] = ["Team", "Total"]  # type: ignore
        home_df.iloc[-2:, 5] = ["Team", "Total"]  # type: ignore
        home_df["Team"] = home_df["Team"].fillna(method="ffill")
        home_df.insert(0, 'Season', season)
        home_df.insert(1, 'Gamecode', gamecode)
        home_df.insert(2, "Home", 1)

        away_df = pd.concat([
            pd.json_normalize(data[1]["PlayersStats"]),
            pd.json_normalize(data[1]["tmr"]),
            pd.json_normalize(data[1]["totr"])
        ])
        away_df.reset_index(drop=True, inplace=True)
        away_df.iloc[-2:, 0] = ["Team", "Total"]  # type: ignore
        away_df.iloc[-2:, 5] = ["Team", "Total"]  # type: ignore
        away_df["Team"] = away_df["Team"].fillna(method="ffill")
        away_df.insert(0, 'Season', season)
        away_df.insert(1, 'Gamecode', gamecode)
        away_df.insert(2, "Home", 0)

        df = pd.concat([home_df, away_df], axis=0)
        return df

    def get_game_boxscore_quarter_data_single_season(
        self,
        season: int,
        boxscore_type: str = "ByQuarter"
    ) -> pd.DataFrame:
        """
        A function that gets the boxscore quarter data of *all* games in a
        single season

        Args:

            season (int): The start year of the season

            boxscore_type (str): The type of quarter boxscore data.
            Available values:
                - ByQuarter
                - EndOfQuarter
            Default: ByQuarter

        Returns:

            pd.DataFrame: A dataframe with the boxscore quarter data of all
                games in a single season
        """
        get_game_boxscore_quarter_data_ = (
            lambda season, gamecode: self.get_game_boxscore_quarter_data(
                season, gamecode, boxscore_type)
        )
        data_df = self.get_season_data_from_game_data(
            season, get_game_boxscore_quarter_data_)
        return data_df

    def get_game_boxscore_quarter_data_multiple_seasons(
        self,
        start_season: int,
        end_season: int,
        boxscore_type: str = "ByQuarter"
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data of *all* games in a range of
        seasons

        Args:

            start_season (int): The start year of the start season

            end_season (int): The start year of the end season

            boxscore_type (str): The type of quarter boxscore data.
                Available values:
                - ByQuarter
                - EndOfQuarter
                Default: ByQuarter

        Returns:

            pd.DataFrame: A dataframe with the boxscore quarter data of all
                games in range of seasons
        """
        get_game_boxscore_quarter_data_ = (
            lambda season, gamecode: self.get_game_boxscore_quarter_data(
                season, gamecode, boxscore_type)
        )
        df = self.get_range_seasons_data(
            start_season, end_season, get_game_boxscore_quarter_data_)
        return df

    def get_player_boxscore_stats_single_season(
        self,
        season: int
    ) -> pd.DataFrame:
        """
        A function that return the player boxscore stats for all games in a
        single season

        Args:
            season (int): The start year of the start season

        Returns:
            pd.DataFrame: A dataframe with home and away team player stats for
                a season
        """
        data_df = self.get_season_data_from_game_data(
            season, self.get_player_boxscore_stats_data)
        return data_df

    def get_player_boxscore_stats_multiple_seasons(
        self,
        start_season: int,
        end_season: int
    ) -> pd.DataFrame:
        """
        A function that return the player boxscore stats for all games in
        multiple season

        Args:
            start_season (int): The start year of the start season

            end_season (int): The start year of the end season

        Returns:
            pd.DataFrame: A dataframe with home and away team player stats for
                a season
        """
        data_df = self.get_range_seasons_data(
            start_season, end_season, self.get_player_boxscore_stats_data)
        return data_df
